"""This module contains logic to interleave a computation graph (an intervention graph) with the computation graph of a model.

The :class:`InterventionProxy <nnsight.intervention.InterventionProxy>` class extends the functionality of a base nnsight.fx.Proxy.Proxy object and makes it easier for users to interact with.

:func:`intervene() <nnsight.intervention.intervene>` is the entry hook into the models computation graph in order to interleave an intervention graph.

The :class:`HookModel <nnsight.intervention.HookModel>` provides a context manager for adding input and output hooks to modules and removing them upon context exit.
"""
from __future__ import annotations

from contextlib import AbstractContextManager
from typing import Any, Callable, Collection, List, Tuple, Union

import torch
from torch.utils.hooks import RemovableHandle

from . import util
from .tracing.Graph import Graph
from .tracing.Proxy import Proxy


class InterventionProxy(Proxy):

    """Sub-class for Proxy that adds additional user functionality to proxies.

    Examples:

        Saving a proxy so it is not deleted at the completion of it's listeners is enabled with ``.save()``:

        .. code-block:: python

            with runner.invoke('The Eiffel Tower is in the city of') as invoker:
                hidden_states = model.lm_head.input.save()
                logits = model.lm_head.output.save()

            print(hidden_states.value)
            print(logits.value)

        This works and would output the inputs and outputs to the model.lm_head module.
        Had you not called .save(), calling .value would have been None.

        Indexing by token of hidden states can easily done using ``.token[<idx>]`` or ``.t[<idx>]``

        .. code-block:: python

            with runner.invoke('The Eiffel Tower is in the city of') as invoker:
                logits = model.lm_head.output.t[0].save()

            print(logits.value)

        This would save only the first token of the output for this module.
        This should be used when using multiple invokes as the batching and padding of multiple inputs could mean the indices for tokens shifts around and this take care of that.

        Calling ``.shape`` on an InterventionProxy returns the shape or collection of shapes for the tensors traced through this module.

        Calling ``.value`` on an InterventionProxy returns the actual populated values, updated during actual execution of the model.

    """

    def save(self) -> InterventionProxy:
        """Method when called, indicates to the intervention graph to not delete the tensor values of the result.

        Returns:
            InterventionProxy: Save proxy.
        """

        # Add a 'null' node with the save proxy as an argument to ensure the values are never deleted.
        # This is because 'null' nodes never actually get set and therefore there will always be a
        # dependency for the save proxy.
        self.node.graph.add(
            value=None,
            target="null",
            args=[self.node],
        )

        return self

    def retain_grad(self):
        self.node.graph.add(target=torch.Tensor.retain_grad, args=[self.node])

        # We need to set the values of self to values of self to add this into the computation graph so grad flows through it
        # This is because in intervene(), we call .narrow on activations which removes it from the grad path
        self.node.graph.add(target=Proxy.proxy_update, args=[self.node, self.node])

    @property
    def token(self) -> TokenIndexer:
        """Property used to do token based indexing on a proxy.
        Directly indexes the second dimension of tensors.
        Makes positive indices negative as tokens are padded on the left.

        Example:

            .. code-block:: python

                model.transformer.h[0].mlp.output.token[0]

            Is equivalent to:

            .. code-block:: python

                model.transformer.h[0].mlp.output.token[:,-3]

            For a proxy tensor with 3 tokens.

        Returns:
            TokenIndexer: Object to do token based indexing.
        """
        return TokenIndexer(self)

    @property
    def t(self) -> TokenIndexer:
        """Property as alias for InterventionProxy.token"""
        return self.token

    @property
    def shape(self) -> Union[torch.Size, Collection[torch.Size]]:
        """Property to retrieve the shape of the traced proxy value.

        Returns:
            Union[torch.Size,Collection[torch.Size]]: Proxy value shape or collection of shapes.
        """
        return util.apply(self.node.proxy_value, lambda x: x.shape, torch.Tensor)

    @property
    def value(self) -> Any:
        """Property to return the value of this proxy's node.

        Returns:
            Any: The stored value of the proxy, populated during execution of the model.
        """

        return self.node.value


def intervene(activations: Any, module_path: str, graph: Graph, key: str):
    """Entry to intervention graph. This should be hooked to all modules involved in the intervention graph.

    Forms the current module_path key in the form of <module path>.<output/input>.<graph generation index>
    Checks the graphs argument_node_names attribute for this key.
    If exists, value is a list of node names to iterate through.
    Node args for argument type nodes should be ``[module_path, batch_size, batch_start]``.
    Using batch_size and batch_start, apply torch.narrow to tensors in activations to select
    only batch indexed tensors relevant to this intervention node. Sets the value of a node
    using the indexed values. Using torch.narrow returns a view of the tensors as opposed to a copy allowing
    subsequent downstream nodes to make edits to the values only in the relevant tensors, and have it update the original
    tensors. This both prevents interventions from effecting bathes outside their preview and allows edits
    to the output from downstream intervention nodes in the graph.

    Args:
        activations (Any): Either the inputs or outputs of a torch module.
        module_path (str): Module path of the current relevant module relative to the root model.
        graph (Graph): Intervention graph to interleave with the computation "graph" of the model.
        key (str): Key denoting either "input" or "output" of module.

    Returns:
        Any: The activations, potentially modified by the intervention graph.
    """

    # Key to module activation argument nodes has format: <module path>.<output/input>.<generation index>
    module_path = f"{module_path}.{key}.{graph.generation_idx}"

    if module_path in graph.argument_node_names:
        argument_node_names = graph.argument_node_names[module_path]

        # multiple argument nodes can have same module_path if there are multiple invocations.
        for argument_node_name in argument_node_names:
            node = graph.nodes[argument_node_name]

            # args for argument nodes are (module_path, batch_size, batch_start)
            _, batch_size, batch_start = node.args

            # We set its result to the activations, indexed by only the relevant batch idxs.
            node.set_value(
                util.apply(
                    activations,
                    lambda x: x.narrow(0, batch_start, batch_size),
                    torch.Tensor,
                )
            )
    return activations


class HookModel(AbstractContextManager):
    """Context manager that applies input and/or output hooks to modules in a model.

    Registers provided hooks on __enter__ and removes them on __exit__.

    Attributes:
        model (torch.nn.Module): Root model to access modules and apply hooks to.
        modules (List[Tuple[torch.nn.Module, str]]): List of modules to apply hooks to along with their module_path.
        input_hook (Callable): Function to apply to inputs of designated modules.
            Should have signature of [inputs(Any), module_path(str)] -> inputs(Any)
        output_hook (Callable): Function to apply to outputs of designated modules.
            Should have signature of [outputs(Any), module_path(str)] -> outputs(Any)
        handles (List[RemovableHandle]): Handles returned from registering hooks as to be used when removing hooks on __exit__.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        module_keys: List[str],
        input_hook: Callable = None,
        output_hook: Callable = None,
        backward_input_hook: Callable = None,
        backward_output_hook: Callable = None,
    ) -> None:
        self.model = model
        self.module_keys = module_keys

        self.input_hook = input_hook
        self.output_hook = output_hook
        self.backward_input_hook = backward_input_hook
        self.backward_output_hook = backward_output_hook

        self.handles: List[RemovableHandle] = []

    def __enter__(self) -> HookModel:
        """Registers input and output hooks to modules if they are defined.

        Returns:
            HookModel: HookModel object.
        """

        for module_key in self.module_keys:
            *module_atoms, hook_type = module_key.split(".")[:-1]
            module_path = ".".join(module_atoms)

            module: torch.nn.Module = util.fetch_attr(self.model, module_path)

            if hook_type == "input":

                def input_hook(module, input, module_path=module_path):
                    return self.input_hook(input, module_path)

                self.handles.append(module.register_forward_pre_hook(input_hook))

            elif hook_type == "output":

                def output_hook(module, input, output, module_path=module_path):
                    return self.output_hook(output, module_path)

                self.handles.append(module.register_forward_hook(output_hook))

            elif hook_type == "backward_input":

                def backward_input_hook(module, input, output, module_path=module_path):
                    return self.backward_input_hook(input, module_path)

                self.handles.append(
                    module.register_full_backward_hook(backward_input_hook)
                )

            elif hook_type == "backward_output":

                def backward_output_hook(module, output, module_path=module_path):
                    return self.backward_output_hook(output, module_path)

                self.handles.append(
                    module.register_full_backward_pre_hook(backward_output_hook)
                )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Removes all handles added during __enter__."""
        for handle in self.handles:
            handle.remove()


class TokenIndexer:
    """Helper class to directly access token indices of hidden states.
    Directly indexes the second dimension of tensors.
    Makes positive indices negative as tokens are padded on the left.

    Args:
        proxy (InterventionProxy): Proxy to aid in token indexing.
    """

    def __init__(self, proxy: InterventionProxy) -> None:
        self.proxy = proxy

    def convert_idx(self, idx: int):
        if idx >= 0:
            n_tokens = self.proxy.node.proxy_value.shape[1]
            idx = -(n_tokens - idx)

        return idx

    def __getitem__(self, key: int) -> Proxy:
        key = self.convert_idx(key)

        return self.proxy[:, key]

    def __setitem__(self, key: int, value: Union[Proxy, Any]) -> None:
        key = self.convert_idx(key)

        self.proxy[:, key] = value



from __future__ import annotations

import collections
from typing import Any, Callable, Dict, List, Tuple, Union

import torch
from torch.utils.hooks import RemovableHandle
from transformers import (AutoConfig, AutoModelForCausalLM, AutoTokenizer,
                          BatchEncoding, PretrainedConfig, PreTrainedModel,
                          PreTrainedTokenizer)

from .AbstractModel import AbstractModel


class LanguageModel(AbstractModel):
    """LanguageModels are nnsight wrappers around AutoModelForCausalLM models.

    Inputs can be in the form of:
        Prompt: (str)
        Prompts: (List[str])
        Batched prompts: (List[List[str]])
        Tokenized prompt: (Union[List[int], torch.Tensor])
        Tokenized prompts: (Union[List[List[int]], torch.Tensor])
        Direct input: (Dict[str,Any])

    If using a custom model, you also need to provide the tokenizer like ``LanguageModel(custom_model, tokenizer=tokenizer)``

    Calls to generate pass arguments downstream to :func:`AutoModelForCausalLM.generate`

    Attributes:
        config (PretrainedConfig): Huggingface config file loaded from repository or checkpoint.
        tokenizer (PreTrainedTokenizer): Tokenizer for LMs.
        meta_model (PreTrainedModel): Meta version of underlying AutoModelForCausalLM model.
        local_model (PreTrainedModel): Local version of underlying AutoModelForCausalLM model.

    """

    def __init__(self, *args, tokenizer=None, **kwargs) -> None:
        self.config: PretrainedConfig = None
        self.tokenizer: PreTrainedTokenizer = tokenizer
        self.meta_model: PreTrainedModel = None
        self.local_model: PreTrainedModel = None

        super().__init__(*args, **kwargs)

    def _register_increment_hook(self, hook: Callable) -> RemovableHandle:
        return self.local_model.register_forward_hook(hook)

    def _load_meta(self, repoid_or_path, *args, **kwargs) -> PreTrainedModel:
        self.config = AutoConfig.from_pretrained(repoid_or_path, *args, **kwargs)

        self.tokenizer = AutoTokenizer.from_pretrained(
            repoid_or_path, config=self.config, padding_side="left"
        )
        self.tokenizer.pad_token = self.tokenizer.eos_token

        return AutoModelForCausalLM.from_config(self.config, trust_remote_code=True)

    def _load_local(self, repoid_or_path, *args, **kwargs) -> PreTrainedModel:
        return AutoModelForCausalLM.from_pretrained(
            repoid_or_path, *args, config=self.config, **kwargs
        )

    def _prepare_inputs(
        self,
        inputs: Union[
            str,
            List[str],
            List[List[str]],
            List[int],
            List[List[int]],
            torch.Tensor,
            Dict[str, Any],
        ],
    ) -> BatchEncoding:
        if isinstance(inputs, collections.abc.Mapping):
            return BatchEncoding(inputs)

        if isinstance(inputs, str) or (
            isinstance(inputs, list) and isinstance(inputs[0], int)
        ):
            inputs = [inputs]

        if isinstance(inputs, torch.Tensor) and inputs.ndim == 1:
            inputs = inputs.unsqueeze(0)

        if not isinstance(inputs[0], str):
            inputs = [{"input_ids": ids} for ids in inputs]

            return self.tokenizer.pad(inputs, return_tensors="pt")

        return self.tokenizer(inputs, return_tensors="pt", padding=True)

    def _batched_inputs(self, prepared_inputs: BatchEncoding) -> torch.Tensor:
        return prepared_inputs["input_ids"]

    def _example_input(self) -> Dict[str, torch.Tensor]:
        return {"input_ids": torch.tensor([[0]])}

    def _scan(self, prepared_inputs, *args, **kwargs) -> None:
        # TODO
        # Actually use args and kwargs. Dont do this now because the args may be specific to _generation which throws unused args errors
        # Maybe inspect signature and filter out unused args.
        self.meta_model(**prepared_inputs.copy().to("meta"))

    def _forward(self, prepared_inputs, *args, **kwargs) -> Any:
        return self.local_model(
            *args, **prepared_inputs.to(self.local_model.device), **kwargs
        )

    def _generation(
        self, prepared_inputs, *args, max_new_tokens: int = 1, **kwargs
    ) -> Any:
        return self.local_model.generate(
            *args,
            **prepared_inputs.to(self.local_model.device),
            max_new_tokens=max_new_tokens,
            **kwargs,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from .Invoker import Invoker
from .Runner import Runner

if TYPE_CHECKING:
    from ..models.AbstractModel import AbstractModel


class DirectInvoker(Runner, Invoker):
    def __init__(
        self, model: "AbstractModel", *args, fwd_args: Dict[str, Any] = None, **kwargs
    ):
        if fwd_args is None:
            fwd_args = dict()

        self.fwd_args = fwd_args

        Runner.__init__(self, model, **self.fwd_args)

        Invoker.__init__(self, self, *args, **kwargs)

    def __enter__(self) -> DirectInvoker:

        Invoker.__enter__(self)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        Invoker.__exit__(self, exc_type, exc_val, exc_tb)
        self.fwd_args.pop("validate", None)
        self.kwargs = self.fwd_args
        self.args = []

        Runner.__exit__(self, exc_type, exc_val, exc_tb)



from __future__ import annotations

import inspect
from typing import Any, Callable, Dict, List, Type, Union
import torch
from .. import util
from ..patching import Patch, Patcher
from .Node import Node
from .Proxy import Proxy, proxy_wrapper


class Graph:
    """Represents a computation graph involving a torch.nn.module.

    Reserved target names:
    
    * 'module' : There should only be the single root module as a node in the graph for tracing. Added on __init__ and when compiling, the node's value is set to to be whatever module that is being interleaved with this computation graph.
    * 'argument' : There can be multiple argument nodes. Their first argument needs to be the argument name which acts as a key in graph.argument_node_names which maps to a list of names for nodes that depend on it's value. These nodes values need to be set outside of the computation graph as entry points to kick of the execution of the graph.
    * 'rtn' : Should only be one 'rtn' target named node as this is what is used.
    * 'null' : Null nodes never get executed and therefore their listeners never get destroyed.

    Attributes:
        validate (bool): If to execute nodes as they are added with their proxy values in order to check if the executions are possible (i.e shape errors etc). Defaults to True.
        proxy_class (Type[Proxy]): Proxy class to use. Defaults to Proxy.
        nodes (Dict[str, Node]): Mapping of node name to node.
        name_idx (Dict[str, int]): Mapping of node target_name to number of previous names with the same target_name.
            Used so names are unique.
        module_proxy (Proxy): Proxy for given root meta module.
        argument_node_names (Dict[str, List[str]]): Map of name of argument to name of nodes that depend on it.
        generation_idx (int): Current generation index.
    """

    @staticmethod
    def trace(
        module: torch.nn.Module, *args: List[Any], **kwargs: Dict[str, Any]
    ) -> Graph:
        """Given a module and some default (should be meta tensors) arguments, create a graph from the module's
        forward method.

        Args:
            module (torch.nn.Module): _description_
            args (List[Any]): desc
            kwargs (Dict[str, Any]): desc

        Returns:
            Graph: _description_
        """

        # Create a graph with the module as the root module
        graph = Graph(module)

        # Get 'unbound' version of forward method so we can pass in proxy of module instead of self
        forward = module.__class__.forward

        # Want list not tuple
        args = list(args)

        # Inspect forward signature to collect all parameters
        signature = inspect.signature(forward)

        def get_argument_value(param: inspect.Parameter, idx: int):
            """Gets the correct argument to pass to forward method.


            Args:
                param (_type_): _description_
                idx (_type_): _description_

            Returns:
                _type_: _description_
            """

            # If idx in range of provided args, create a proxy for that arg instead of default.
            if idx < len(args):
                return graph.add(
                    graph=graph, value=args[idx], target="argument", args=[param.name]
                )
            # If param name in provided kwargs, create a proxy for that arg instead of default.
            if param.name in kwargs:
                return graph.add(
                    graph=graph,
                    value=kwargs[param.name],
                    target="argument",
                    args=[param.name],
                )
            # Otherwise just return default
            return param.default

        # Create the appropriate proxies/values for the forward method in order to trace.
        arguments = [
            get_argument_value(param, i)
            for i, param in enumerate(list(signature.parameters.values())[1:])
        ]

        # Some methods cannot be caught because they aren't torch functions or dont play nice with __torch_function__.
        # So the patcher replaces the methods with something to catch proxies and return proxies.
        with Patcher() as patcher:
            patcher.add(Patch(torch.full, proxy_wrapper(torch.full)))
            patcher.add(Patch(torch.finfo, proxy_wrapper(torch.finfo)))
            patcher.add(Patch(torch.arange, proxy_wrapper(torch.arange)))

            # Run forward with root module proxy and arguments
            output: Proxy = forward(graph.module_proxy, *arguments)

            # Get proxy_value for return
            value = util.apply(output, lambda x: x.node.proxy_value, Proxy)

            # Create the 'rtn_0' return proxy
            return_proxy = graph.add(
                graph=graph, value=value, target=Graph.rtn, args=output
            )

            # This is how we tell the graph not to destroy a proxy after it's listeners are completed.
            # Create a 'null' proxy. The return proxy listens to the 'null' proxy with args=[return_proxy.node] but 'null' will never be completed.
            graph.add(
                graph=graph,
                value=None,
                target="null",
                args=[return_proxy.node],
            )

        return graph

    @staticmethod
    def rtn(*args, **kwargs):
        """
        Function to just pass through data for returning data in a graph forward method.

        Returns:
            _type_: _description_
        """

        return args

    def __init__(
        self,
        module: torch.nn.Module,
        proxy_class: Type[Proxy] = Proxy,
        validate: bool = True,
    ) -> None:
        self.proxy_class = proxy_class
        self.validate = validate

        self.nodes: Dict[str, Node] = dict()
        self.name_idx: Dict[str, int] = dict()

        self.module_proxy = self.add(value=module, target="module")
        self.argument_node_names: Dict[str, List[str]] = dict()

        self.generation_idx = 0

    def increment(self) -> None:
        """Increments the generation_idx by one. Should be called by a forward hook on the model being used for generation."""
        self.generation_idx += 1

    def compile(self, module: torch.nn.Module) -> None:
        """Re-compile graph to prepare for a new execution of the graph.

        Compiles all nodes and sets generation_idx to 0.

        Finally, sets the "module_0" node's value to the module that is being interleaved.

        Args:
            module (torch.nn.Module): Module to be considered the root module of the graph.
        """

        # Remove nodes that have no effect.
        self.eliminate_dead_code()

        # Compile nodes individually.
        for node in self.nodes.values():
            node.compile()

        self.generation_idx = 0

        # Setting the root module kicks off the graph execution.
        self.nodes["module_0"].set_value(module)

    def add(
        self,
        target: Union[Callable, str],
        value: Any = inspect._empty,
        args: List[Any] = None,
        kwargs: Dict[str, Any] = None,
        name: str = None,
    ) -> Proxy:
        """Adds a node to the graph and returns it's proxy.

        Args:
            value (Any): 'meta' proxy value used for tracing the shapes and values.
            target (Union[Callable, str]): Either the function to call for this node, or a string of a reserved target name.
            args (List[Any], optional): Positional arguments of node. Defaults to None.
            kwargs (Dict[str, Any], optional): Keyword arguments of node. Defaults to None.
            name (str, optional): Unique name of node. Otherwise pull name from target Defaults to None.

        Returns:
            Proxy: Proxy for the added node.

        Raises:
            ValueError: If more than one reserved "rtn" or "module" nodes are added to the graph.
        """

        # If we're validating and the user did not provide a value, execute the given target with meta proxy values to compute new proxy_value.
        if self.validate and value is inspect._empty:
            _args = args if args is not None else []
            _kwargs = kwargs if kwargs is not None else {}

            value = target(
                *Node.prepare_proxy_values(_args),
                **Node.prepare_proxy_values(_kwargs),
            )

        target_name = target if isinstance(target, str) else target.__name__

        if target_name not in self.name_idx:
            self.name_idx[target_name] = 0
        else:
            if target_name == "rtn":
                raise ValueError("Can only have one return ('rtn') node.")
            if target_name == "module":
                raise ValueError("Can only have one module node.")

        if name is None:
            name = f"{target_name}_{self.name_idx[target_name]}"

        self.name_idx[target_name] += 1

        stack = inspect.stack()
        proxy_frame = stack[2]

        node = Node(
            name=name,
            graph=self,
            value=value,
            target=target,
            args=args,
            kwargs=kwargs,
            meta={"line": proxy_frame.lineno, "file": proxy_frame.filename},
        )

        self.nodes[name] = node

        if target_name == "argument":
            module_path = args[0]

            if module_path not in self.argument_node_names:
                self.argument_node_names[module_path] = []

            self.argument_node_names[module_path].append(name)

        return self.proxy(node)

    def proxy(self, node: Node) -> Proxy:
        """Returns proxy of node with specified proxy_class.

        Args:
            node (Node): Node.

        Returns:
            Proxy: Proxy.
        """
        return self.proxy_class(node)

    def eliminate_dead_code(self):
        # TODO
        pass

    def wrap(self, module: torch.nn.Module) -> torch.nn.Module:
        """Replaces the forward method of the given module with an execution of the module's graph.

        Args:
            module (torch.nn.Module): Module to replace the forward method of.

        Returns:
            torch.nn.Module: The module, post-replacement.
        """

        def forward(*args, **kwargs):
            # Compile the graph with the given module as the root module.
            self.compile(module)

            # Gets list of all argument nodes for this graph.
            argument_nodes_list = list(self.argument_node_names.values())

            # Sets the result of the argument nodes for args.
            for i, arg in enumerate(args):
                self.nodes[argument_nodes_list[i][0]].set_value(arg)

            # And then for kwargs.
            for key in kwargs:
                if key in self.argument_node_names:
                    self.nodes[self.argument_node_names[key][0]].set_value(arg)

            # 'rtn_0' should have the value we need to return.
            return_value = self.nodes["rtn_0"].value
            self.nodes["rtn_0"].destroy()
            return return_value

        # Replace forward method with custom graph execution method.
        module.forward = forward

        return module

    def __str__(self) -> str:
        result = ""

        for name, node in self.nodes.items():
            result += f"  %{node}\n"

        return result



# The library is called nnsight
from nnsight import LanguageModel
import torch
# Get model wrapper for any model you can get with AutoConfig.from_pretrained(model_name)
model = LanguageModel('gpt2',device_map='cuda:0')

# Prints normal module tree to show access tree for modules
print(model)

# Invoke using a prompt
with model.generate(max_new_tokens=3) as generator:
    with generator.invoke('Hello world ') as invoker:

        # See the input prompt seperated into token strings
        tokenized = invoker.tokens

        # Use normal module access and .output to get output activations.
        # Then save the activations at this point in the execution tree
        # not only to retrieve later on (by calling mlp0.value after the model has been ran),
        # but if this value changes throughout interventions and you want the value before those alterations.

        mlp0  = model.transformer.h[0].mlp.output.save()

        # Use .token[idx] or .t[idx] to index by token
        mlp0_t1_t = model.transformer.h[0].mlp.output.t[0].save()

        mmlp0 = model.transformer.h[0].mlp.output
        mmlp1 = model.transformer.h[1].mlp.output 
        # Addition works like you normally would either with tensors or primatives ( will add other operations later)
        noise = (0.001**0.5)*torch.randn(mmlp1.t[1].shape)
        mmlp1 = mmlp1.t[1] + noise

        mmlp2_before = model.transformer.h[2].mlp.output.save()

        # Easily set the output of a module to whatever you want
        model.transformer.h[2].mlp.output = mmlp0 + mmlp1

        # See the before and after the intervation
        mmlp2_after = model.transformer.h[2].mlp.output.save()

    with generator.invoke('Goodbye world') as invoker:

        # Operations work cross-prompt!
        model.transformer.h[1].mlp.output = mmlp0




from nnsight import LanguageModel

model = LanguageModel('decapoda-research/llama-65b-hf')

print(model)

with model.generate(server=True, max_new_tokens=1, return_dict_in_generate=True, output_scores=True) as generator:
    with generator.invoke('Hello world') as invoker:
        
        hiddenstates = model.model.layers[0].output.save()

print(hiddenstates.value)



from nnsight import LanguageModel
import objgraph
import random
import gc
import torch
model = LanguageModel('EleutherAI/pythia-70m', device_map='cuda:0')



i = 0

with torch.no_grad():
    while True:

        with model.forward("sadas sadasjasd", scan=False):
            pass



from __future__ import annotations

from contextlib import AbstractContextManager
from typing import Any, Dict

from ..tracing.Proxy import Proxy
from .Tracer import Tracer


class Invoker(AbstractContextManager):
    """An Invoker is meant to work in tandem with a :class:`nnsight.contexts.Tracer.Tracer` to enter input and manage intervention tracing.

    Attributes:
        tracer (nnsight.contexts.Tracer.Tracer): Tracer object to enter input and manage context.
        input (Any): Initially entered input, then post-processed input from model's _prepare_inputs method.
        scan (bool): If to use a 'meta' version of the  post-processed input to run through the model using it's _scan method,
            in order to update the potential sizes/dtypes of all module's inputs/outputs as well as validate things work correctly.
            Scanning is not free computation wise so you may want to turn this to false when running in a loop.
            When making interventions, you made get shape errors if scan is false as it validates operations based on shapes so
            for looped calls where shapes are consistent, you may want to have scan=True for the first loop. Defaults to True.
        args (List[Any]): Positional arguments passed to the model's _prepare_inputs method.
        kwargs (Dict[str,Any]): Keyword arguments passed to the model's _prepare_inputs method.
    """

    def __init__(
        self,
        tracer: Tracer,
        input: Any,
        *args,
        scan: bool = True,
        **kwargs,
    ) -> None:
        self.tracer = tracer
        self.input = input
        self.scan = scan
        self.args = args
        self.kwargs = kwargs

    def __enter__(self) -> Invoker:
        """Enters a new invocation context with a given input.

        Sets the generation_idx to 0.
        Calls the model's _prepare_inputs method using the input and other arguments.
        If scan is True, uses the model's _scan method to update and validate module inputs/outputs.
        Gets a batched version of the post processed input using the model's _batched_inputs method to update the Tracer's
            current batch_size and batched_input.

        Returns:
            Invoker: Invoker.
        """
        # Were in a new invocation so set generation_idx to 0,
        self.tracer.generation_idx = 0

        self.input = self.tracer.model._prepare_inputs(
            self.input, *self.args, **self.kwargs
        )

        if self.scan:
            self.tracer.model._scan(self.input, *self.tracer.args, **self.tracer.kwargs)
        else:
            for name, module in self.tracer.model.meta_model.named_modules():
                module._output = None
                module._input = None
                module._backward_output = None
                module._backward_input = None

        batched_inputs = self.tracer.model._batched_inputs(self.input)

        self.tracer.batch_size = len(batched_inputs)
        self.tracer.batched_input.extend(batched_inputs)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if isinstance(exc_val, BaseException):
            raise exc_val

    def next(self, increment: int = 1) -> None:
        """Designates subsequent interventions should be applied to the next generation for multi-iteration generation runs.

        Args:
            increment (int): How many generation_idx to increment at once. Defaults to 1.
        """
        # .next() increases which generation idx the interventions happen.
        self.tracer.generation_idx += increment

        if self.scan:
            # Run graph with singe token input.
            self.inputs = self.tracer.model._prepare_inputs(
                self.tracer.model._example_input(), *self.args, **self.kwargs
            )
            self.tracer.model._scan(
                self.inputs, *self.tracer.args, **self.tracer.kwargs
            )
        else:
            for name, module in self.tracer.model.meta_model.named_modules():
                module._output = None
                module._input = None
                module._backward_output = None
                module._backward_input = None

    def save_all(self) -> Dict[str, Proxy]:
        """Saves the output of all modules and returns a dictionary of [module_path -> save proxy]

        Returns:
            Dict[str, Proxy]: Dictionary of all modules saved, keyed by their module_path.
        """
        result = {}

        for name, module in self.tracer.model.meta_model.named_modules():
            result[module.module_path] = module.output.save()

        return result



"""This module contains the main Model classes which enable the tracing and interleaving functionality of nnsight. 

Models allow users to load and wrap torch modules. Here we load gpt2 from HuggingFace using its repo id:

.. code-block:: python

    from nnsight import LanguageModel
    model = LanguageModel('gpt2', device_map='cuda:0')

In this case, declaring a LanguageModel entails the underlying model is a ``transformers.AutoModelForCausalLM``, and unused arguments by LanguageModel are passed downstream to AutoModelForCausalLM. ``device_map='cuda:0'`` leverages the accelerate package to use the first GPU when loading the local model.

The wrapping of the underlying model encompasses both the display of its module structure when the model is printed and the ability to directly access the attributes of the underlying meta_model through the wrapped model itself.

Printing out the wrapped module returns its structure:

.. code-block:: python

    print(model)

.. code-block:: text

    GPT2LMHeadModel(
    (transformer): GPT2Model(
        (wte): Embedding(50257, 768)
        (wpe): Embedding(1024, 768)
        (drop): Dropout(p=0.1, inplace=False)
        (h): ModuleList(
        (0-11): 12 x GPT2Block(
            (ln_1): LayerNorm((768,), eps=1e-05, elementwise_affine=True)
            (attn): GPT2AttentionAltered(
            (c_attn): Conv1D()
            (c_proj): Conv1D()
            (attn_dropout): Dropout(p=0.1, inplace=False)
            (resid_dropout): Dropout(p=0.1, inplace=False)
            (query): WrapperModule()
            (key): WrapperModule()
            (value): WrapperModule()
            )
            (ln_2): LayerNorm((768,), eps=1e-05, elementwise_affine=True)
            (mlp): GPT2MLP(
            (c_fc): Conv1D()
            (c_proj): Conv1D()
            (act): NewGELUActivation()
            (dropout): Dropout(p=0.1, inplace=False)
            )
        )
        )
        (ln_f): LayerNorm((768,), eps=1e-05, elementwise_affine=True)
    )
    (lm_head): Linear(in_features=768, out_features=50257, bias=False)
    )

The primary methods of interacting and running the model are ``.generate(...)`` and ``.forward(...)``. Both return context manager objects which, when entered, track operations performed on the inputs and outputs of modules.

The :func:`generate <nnsight.models.AbstractModel.AbstractModel.generate>` context is meant for multi-iteration runs. Arguments passed to generate determine the generation behavior — in this case to generate three tokens.
Within a generation context, invoker sub-contexts are entered using ``generator.invoke``. This is where an input (or batch of inputs) to the model is accepted, and batched with other invocations. It's in these contexts where operations on inputs and outputs of modules are tracked and prepared for execution.

In this example, we run two prompts on the language model in order to generate three tokens. We also perform a ``.save()`` operation on the output of the lm_head module (the logit outputs) in order to save these activations and access them after generation is over:

.. code-block:: python

    with model.generate(max_new_tokens=3) as generator:
        with generator.invoke("The Eiffel Tower is in the city of") as invoker:
            logits1 = model.lm_head.output.save()
        with generator.invoke("The Empire State Building is in the city of") as invoker:
            logits2 = model.lm_head.output.save()

    print(logits1.value)
    print(logits2.value)

The :func:`forward <nnsight.models.AbstractModel.AbstractModel.forward>` context is meant for direct input to the underlying model (or module).

.. code-block:: python

    with model.forward(inference=True) as runner:
        with runner.invoke("The Eiffel Tower is in the city of") as invoker:
            logits = model.lm_head.output.save()

    print(logits.value)

See :mod:`nnsight.contexts` for more.

"""



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple, Union

import torch

from .. import util
from ..logger import logger
from .Proxy import Proxy

if TYPE_CHECKING:
    from .Graph import Graph


class Node:
    """A Node represents some action that should be carried out during execution of a Graph.

    Attributes:
        name (str): Unique name of node.
        graph (Graph): Reference to parent Graph object.
        proxy_value (Any): Meta version of value. Used when graph has validate = True.
        target (Union[Callable, str]): Function to execute or reserved string name.
        args (List[Any], optional): Positional arguments. Defaults to None.
        kwargs (Dict[str, Any], optional): Keyword arguments. Defaults to None.
        meta (Dict[str, Any], optional): Meta information (used when tracing whole modules). Defaults to None.
        listeners (List[Node]): Nodes that depend on this node.
        dependencies (List[Node]): Nodes that this node depends on.
        value (Any): Actual value to be populated during execution.
        _proxy_device (torch.device): desc
    """

    @staticmethod
    def prepare_proxy_values(values):
        """Prepare arguments for validating a node's target.
        Converts Proxies and Nodes to their proxy_value and moves tensors to 'meta' device.

        Args:
            values (Any): Values to prepare.
        Returns:
            values (Any): Prepared values.
        """

        def slice_to_value(arg: slice):
            return slice(
                Node.prepare_proxy_values(arg.start),
                Node.prepare_proxy_values(arg.stop),
                Node.prepare_proxy_values(arg.step),
            )

        # Convert proxies to their proxy_value
        values = util.apply(values, lambda x: x.node.proxy_value, Proxy)
        # Convert nodes to their proxy_value
        values = util.apply(values, lambda x: x.proxy_value, Node)
        # Slices may have proxies as part of their attributes so convert those to their proxy_values
        values = util.apply(values, slice_to_value, slice)
        # Move tensors to 'meta'
        values = util.apply(values, lambda x: x.to("meta"), torch.Tensor)

        return values

    def __init__(
        self,
        name: str,
        graph: "Graph",
        value: Any,
        target: Union[Callable, str],
        args: List[Any] = None,
        kwargs: Dict[str, Any] = None,
        meta: Dict[str, Any] = None,
    ) -> None:
        super().__init__()

        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        if meta is None:
            meta = dict()

        self.name = name
        self.graph = graph
        self.proxy_value = value
        self.target = target
        self.args = util.apply(args, lambda x: x.node, Proxy)
        self.kwargs = util.apply(kwargs, lambda x: x.node, Proxy)
        self.meta = meta

        self.value: Any = None

        self.listeners: List[Node] = list()
        self.dependencies: List[Node] = list()

        # Add all arguments that are nodes to nodes dependencies
        util.apply(self.args, lambda x: self.dependencies.append(x), Node)
        util.apply(self.kwargs, lambda x: self.dependencies.append(x), Node)
        # Add node to all arguments that are nodes' listeners
        util.apply(self.args, lambda x: x.listeners.append(self), Node)
        util.apply(self.kwargs, lambda x: x.listeners.append(self), Node)

        self.remaining_listeners = 0
        self.remaining_dependencies = 0

        self._proxy_device: torch.device = None

    @property
    def proxy_device(self) -> torch.device:
        """Lazy creation of _proxy_device attribute.

        Returns:
            torch.Device: _description_
        """
        if self._proxy_device is None:
            device = None

            def _device(value):
                nonlocal device
                device = value.device

            util.apply(self.proxy_value, _device, torch.Tensor)
            # TODO
            # util.apply(self.proxy_value, _device, torch.nn.Module)

            self._proxy_device = device

        return self._proxy_device

    def compile(self) -> None:
        """Resets this Nodes remaining_listeners and remaining_dependencies and sets its value to None."""
        self.remaining_listeners = len(self.listeners)
        self.remaining_dependencies = len(self.dependencies)
        self.value = None

    def fulfilled(self) -> bool:
        """Returns true if remaining_dependencies is 0.

        Returns:
            bool: If fulfilled.
        """
        return self.remaining_dependencies == 0

    def redundant(self) -> bool:
        """Returns true if remaining_listeners is 0.

        Returns:
            bool: If redundant.
        """
        return self.remaining_listeners == 0

    def prepare_inputs(self) -> Tuple[List[Any], Dict[str, Any]]:
        """Prepare arguments for executing this node's target.
        Converts Nodes in args and kwargs to their value and moves tensors to correct device.


        Returns:
            Tuple[List[Any], Dict[str, Any]]: Prepared args and kwargs
        """

        # Turn nodes into their value
        def _value(node: Node):
            return node.value

        args = util.apply(self.args, _value, Node)
        kwargs = util.apply(self.kwargs, _value, Node)

        device = None

        def _device(value):
            nonlocal device
            device = value.device

        all_args = list(args) + list(kwargs.values())

        util.apply(list(reversed(all_args)), _device, torch.Tensor)
        # TODO
        # util.apply(list(reversed(all_args)), _device, torch.nn.Module)

        # Move tensors to device
        def _to(value: torch.Tensor):
            return value.to(device)

        args = util.apply(args, _to, torch.Tensor)
        kwargs = util.apply(kwargs, _to, torch.Tensor)

        return args, kwargs

    def execute(self) -> None:
        """Actually executes this node.
        If target is 'null' do nothing.
        Prepares args and kwargs and passed them to target.
        """

        # We se a nodes target to 'null' if we don't want it to be executed and therefore never done
        if self.target == "null":
            return

        # Prepare arguments.
        args, kwargs = self.prepare_inputs()

        # Call the target to get value.
        output = self.target(*args, **kwargs)

        self.set_value(output)

    def set_value(self, value: Any):
        """Sets the value of this Node and logs the event.
        Updates remaining_dependencies of listeners. If they are now fulfilled, execute them.
        Updates remaining_listeners of dependencies. If they are now redundant, destroy them.

        Args:
            value (Any): Value.
        """
        self.value = value

        logger.info(f"=> SET({self.name})")

        for listener in self.listeners:
            listener.remaining_dependencies -= 1

            if listener.fulfilled():
                listener.execute()

        for dependency in self.dependencies:
            dependency.remaining_listeners -= 1

            if dependency.redundant():
                dependency.destroy()

        if self.value is not None and self.redundant():
            self.destroy()

    def destroy(self) -> None:
        """Removes the reference to the node's value and logs it's destruction."""
        logger.info(f"=> DEL({self.name})")

        self.value = None

    def __str__(self) -> str:
        args = util.apply(self.args, lambda x: f"'{x}'", str)
        args = util.apply(args, lambda x: x.name, Node)
        args = [str(arg) for arg in args]
        meta = f"{self.meta['file']}({self.meta['line']})" if self.meta else ""
        return f"{self.name}:[ {meta} args:({','.join(args)}) l:{len(self.listeners)} d:{len(self.dependencies)}]"



"""The nnsight.tracing module involves tracking operations on a torch.nn.Module in order to form a computation graph.

The :class:`Graph <nnsight.tracing.Graph.Graph>` class adds and stores each operation or node. It has a 'module' node which acts as the root object on which the computation graph is performing.
It has 'argument' nodes which act as entry points for data to flow into the graph.

The :class:`Node <nnsight.tracing.Node.Node>` class represents nodes in the graph. The class represents the operations (and the resulting output of said operations) they are tracing AND nodes that actually execute the operations when running the graph on a model.

* Nodes have a ``.proxy_value`` attribute that are a result of the tracing operation, and are 'meta' tensors allowing you to view the shape and datatypes of the actual resulting value that will be populated when the node' operation is executed.
* Nodes carry out their operation in ``.execute()`` where their arguments are pre-processed and their value is set in ``.set_value()``.
* Arguments passed to the node are other nodes, where a bi-directional dependency graph is formed. During execution pre-processing, the arguments that are nodes and converted to their value.
* Nodes are responsible for updating their listeners that one of their dependencies are completed, and if all are completed that they should execute. Similarly, nodes must inform their dependencies when one of their listeners has ceased "listening." If the node has no listeners, it's value is destroyed by calling ``.destroy()`` in order to free memory. When re-executing the same graph and therefore the same nodes, the remaining listeners and dependencies are reset on each node.

:class:`Proxy <nnsight.tracing.Proxy.Proxy>` class objects are the actual objects that interact with operations in order to update the graph to create new nodes. Each Node has it's own proxy object. The operations that are traceable on base Proxy objects are many python built-in and magic methods, as well as implementing __torch_function__ to trace torch operations. When an operation is traced, arguments are converted into their 'meta' tensor values and ran through the operation in order to find out the shames and data types of the result.
"""



from nnsight import LanguageModel

model = LanguageModel("gpt2")

print(model)

with model.generate(server=True) as generator:
    with generator.invoke("Hello world") as invoker:
        hiddenstates = model.transformer.h[2].output.save()


print(hiddenstates.value)



"""The contexts module contains logic for managing the tracing and running of models with :mod:`nnsight.tracing` and :mod:`nnsight.module`

The primary two classes involved here are :class:`Tracer <nnsight.contexts.Tracer.Tracer>` and :class:`Invoker <nnsight.contexts.Invoker.Invoker>`.

The :class:`Tracer <nnsight.contexts.Tracer.Tracer>` class creates a :class:`Graph <nnsight.tracing.Graph.Graph>` around the meta_model of an :class:`AbstractModel <nnsight.models.AbstractModel.AbstractModel>`. The graph tracks and manages the operations performed on the inputs and outputs of said model.
Modules in the meta_model expose their ``.output`` and ``.input`` attributes which when accessed, add to the computation graph of the tracer.
To do this, they need to know about the current Tracer object, so each Module's ``.tracer`` object is set to be the current Tracer.

The Tracer object also keeps track of the batch_size of the most recent input, the generation index for multi iteration generations, and all of the inputs made during its context in the ``.batched_input`` attribute. Inputs added to this attribute should be in a format where each index is a batch index and allows the model to batch all of the inputs together.

This is to keep things consistent. If two different inputs are in two different valid formats, they both become the same format and are easy to batch.
In the case of LanguageModels, regardless of whether the input are string prompts, pre-processed dictionaries, or input ids, the batched input is only input ids.
On exiting the Tracer context, the Tracer object should use the information and inputs provided to it to carry out the execution of the model and update its ``.output`` attribute with the result.

The :class:`Invoker <nnsight.contexts.Invoker.Invoker>` class is what actually accepts inputs to the model/graph, and it updates its parent Tracer object with the appropriate information about respective inputs. On entering the invoker context with some input, the invoker leverages the model to pre-process and prepare the input to the model. Using the prepared inputs, it updates its Tracer object with a batched version of the input, the size of the batched input, and the current generation index. It also runs a 'meta' version of the input through the model's meta_model. This updates the sizes/dtypes of all of the Module's inputs and outputs based on the characteristics of the input.

nnsight comes with an implementation of a Tracer which can access a models ``._generation`` or ``._forward`` method.

The :class:`Runner <nnsight.contexts.Runner.Runner>` class is a feature rich context manager built for multi-iteration and multi-input executions of the model.
Each Runner context uses invoker contexts within to add new inputs to the eventual execution like ``generator.invoke(...)``
The Runner object also has logic to perform the execution of the traced computation graph on NDIF remote servers (assuming they are running!).
"""


from nnsight import DiffuserModel


diffuser = DiffuserModel("CompVis/stable-diffusion-v1-4", device='cuda:0')


with diffuser.generate() as generator:

    with generator.invoke(["Blue elephant", "GREEN"]) as invoker:

        pass
diffuser.to_image(generator.output)[0].save('ummb.png')
diffuser.to_image(generator.output)[1].save('umm.png')


from nnsight import LanguageModel

model = LanguageModel("gpt2", device_map='cuda:0')


def get_scores():
    hs = model.transformer.h[-1].output[0]

    return model.lm_head(model.transformer.ln_f(hs))


def decode(scores):
    scores = scores.argmax(dim=2)[0, -1]
    return model.tokenizer.decode(scores)


with model.generate(
    max_new_tokens=3,
    return_dict_in_generate=True,
    output_scores=True,
) as generator:
    with generator.invoke(
        "Madison square garden is located in the city of New"
    ) as invoker:
        tokenized = invoker.tokens

        # Reference the hidden states of the last layer for each token of the nine tokens (shape: (1,9,768))
        # Apply lm_head (decode into vocabulary space) and copy and return value (shape: (1,9,50257))
        logits1 = get_scores().save()

        # Denote that you are generating a token and subsequent interventions will apply to that generation
        # and not the previous ones.
        invoker.next()

        # Here the shape of the hidden states is (1, 1, 768) as there is just the one token
        # Get its hidden states of the last layer decoded as well
        logits2 = get_scores().save()

        # And again....
        invoker.next()

        logits3 = get_scores().save()


pred1 = decode(logits1.value)
pred2 = decode(logits2.value)
pred3 = decode(logits3.value)

print(pred1)
print(pred2)
print(pred3)



from typing import Any

import torch
from nnsight import AbstractModel, LanguageModel, util
from nnsight.module import Module
from nnsight.toolbox.optim.lora import LORA
from torch.utils.data import DataLoader, Dataset

model = LanguageModel("gpt2", device_map="cuda:0")

n_tokens = 10
epochs = 1
answer = "Paris"
answer_tokens = model.tokenizer(answer)
answer_token = answer_tokens["input_ids"][0]

lora = LORA(model.transformer.h[0].mlp, 10)

optimizer = torch.optim.AdamW(lora.parameters(), lr=.1)
dataset = [[" ".join(["_"] * n_tokens), answer_token]] * 100
dataloader = DataLoader(dataset, batch_size=10)


lossfn = util.cross_entropy_loss

for epoch in range(epochs):
    print(epoch)

    for i, (inputs, targets) in enumerate(dataloader):
        print(f"  {i}")

        optimizer.zero_grad()

        with model.forward(inputs, inference=False) as runner:


            lora()

            logits = model.lm_head.output.save()

        print(lora.WA)
        loss = lossfn(logits.value[:, -1], targets)
        print(loss)

        loss.backward()

        optimizer.step()


with model.generate() as generator:
    with generator.invoke(dataset[0][0]) as invoker:
        pass

print(model.tokenizer.decode(generator.output[0]))


with model.generate() as generator:
    with generator.invoke(dataset[0][0]) as invoker:
        lora()

print(model.tokenizer.decode(generator.output[0]))



from nnsight import LanguageModel
from nnsight.tracing.Proxy import Proxy
from nnsight import util
import torch

model = LanguageModel('gpt2', device_map='cuda:0')

clean_prompt = "After John and Mary went to the store, Mary gave a bottle of milk to"
corrupted_prompt = "After John and Mary went to the store, John gave a bottle of milk to"

correct_index = model.tokenizer(" John")['input_ids'][0]
incorrect_index = model.tokenizer(" Mary")['input_ids'][0]


with model.generate(max_new_tokens=1) as generator:

    with generator.invoke(clean_prompt) as invoker:

        clean_tokens = invoker.input['input_ids'][0]

        clean_hs = [model.transformer.h[layer_idx].output[0] for layer_idx in range(len(model.transformer.h))]

        clean_logits = model.lm_head.output

        clean_logit_diff = (clean_logits[0, -1, correct_index] - clean_logits[0, -1, incorrect_index]).save()

    with generator.invoke(corrupted_prompt) as invoker:

        corrupted_logits = model.lm_head.output

        corrupted_logit_diff = (corrupted_logits[0, -1, correct_index] - corrupted_logits[0, -1, incorrect_index]).save()

    ioi_patching_results = []

    for layer_idx in range(len(model.transformer.h)):

        _ioi_patching_results = []

        for token_idx in range(len(clean_tokens)):

            with generator.invoke(corrupted_prompt) as invoker:

                model.transformer.h[layer_idx].output[0].t[token_idx] = clean_hs[layer_idx].t[token_idx]

                patched_logits = model.lm_head.output

                patched_logit_diff = patched_logits[0, -1, correct_index] - patched_logits[0, -1, incorrect_index]

                patched_result = (patched_logit_diff - corrupted_logit_diff)/(clean_logit_diff - corrupted_logit_diff)

                _ioi_patching_results.append(patched_result.save())

        ioi_patching_results.append(_ioi_patching_results)


print(f"Clean logit difference: {clean_logit_diff.value:.3f}")
print(f"Corrupted logit difference: {corrupted_logit_diff.value:.3f}")
breakpoint()
ioi_patching_results = util.apply(ioi_patching_results, lambda x : x.value, Proxy)
ioi_patching_results = util.apply(ioi_patching_results, lambda x : x.item(), torch.Tensor)

import plotly.express as px
clean_tokens = [model.tokenizer.decode(token) for token in clean_tokens]
token_labels = [f"{token}_{index}" for index, token in enumerate(clean_tokens)]

fig = px.imshow(ioi_patching_results, color_continuous_midpoint=0.0, color_continuous_scale="RdBu", labels={"x":"Position", "y":"Layer"}, x=token_labels, title="Normalized Logit Difference After Patching Residual Stream on the IOI Task")

fig.write_image("patching.png")

breakpoint()


from __future__ import annotations

import collections
from typing import Any, Callable, Dict, List, Optional, Union

import diffusers
import torch
from diffusers import DiffusionPipeline, SchedulerMixin
from PIL import Image
from torch.utils.hooks import RemovableHandle
from transformers import BatchEncoding, CLIPTextModel, CLIPTokenizer

from .AbstractModel import AbstractModel


class Diffuser(torch.nn.Module):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

        self.pipeline = DiffusionPipeline.from_pretrained(*args, **kwargs)

        for key, value in self.pipeline.__dict__.items():
            if isinstance(value, torch.nn.Module):
                setattr(self, key, value)

        self.tokenizer = self.pipeline.tokenizer

    @torch.no_grad()
    def scan(
        self,
        prompt: Union[str, List[str]] = None,
        height: Optional[int] = None,
        width: Optional[int] = None,
        num_inference_steps: int = 50,
        guidance_scale: float = 7.5,
        negative_prompt: Optional[Union[str, List[str]]] = None,
        num_images_per_prompt: Optional[int] = 1,
        eta: float = 0.0,
        generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,
        latents: Optional[torch.FloatTensor] = None,
        prompt_embeds: Optional[torch.FloatTensor] = None,
        negative_prompt_embeds: Optional[torch.FloatTensor] = None,
        output_type: Optional[str] = "pil",
        return_dict: bool = True,
        callback: Optional[Callable[[int, int, torch.FloatTensor], None]] = None,
        callback_steps: int = 1,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
        guidance_rescale: float = 0.0,
    ):

        # 0. Default height and width to unet
        height = height or self.pipeline.unet.config.sample_size * self.pipeline.vae_scale_factor
        width = width or self.pipeline.unet.config.sample_size * self.pipeline.vae_scale_factor

        # 1. Check inputs. Raise error if not correct
        self.pipeline.check_inputs(
            prompt, height, width, callback_steps, negative_prompt, prompt_embeds, negative_prompt_embeds
        )

        # 2. Define call parameters
        if prompt is not None and isinstance(prompt, str):
            batch_size = 1
        elif prompt is not None and isinstance(prompt, list):
            batch_size = len(prompt)
        else:
            batch_size = prompt_embeds.shape[0]

        do_classifier_free_guidance = guidance_scale > 1.0

        # 3. Encode input prompt
        text_encoder_lora_scale = (
            cross_attention_kwargs.get("scale", None) if cross_attention_kwargs is not None else None
        )
        prompt_embeds, negative_prompt_embeds = self.pipeline.encode_prompt(
            prompt,
            'meta',
            num_images_per_prompt,
            do_classifier_free_guidance,
            negative_prompt,
            prompt_embeds=prompt_embeds,
            negative_prompt_embeds=negative_prompt_embeds,
            lora_scale=text_encoder_lora_scale,
        )
        # For classifier free guidance, we need to do two forward passes.
        # Here we concatenate the unconditional and text embeddings into a single batch
        # to avoid doing two forward passes
        if do_classifier_free_guidance:
            prompt_embeds = torch.cat([negative_prompt_embeds, prompt_embeds])

        # 4. Prepare timesteps
        timesteps = self.pipeline.scheduler.timesteps

        # 5. Prepare latent variables
        num_channels_latents = self.pipeline.unet.config.in_channels
        latents = self.pipeline.prepare_latents(
            batch_size * num_images_per_prompt,
            num_channels_latents,
            height,
            width,
            prompt_embeds.dtype,
            'meta',
            generator,
            latents,
        )

        # 6. Prepare extra step kwargs. TODO: Logic should ideally just be moved out of the pipeline
        extra_step_kwargs = self.pipeline.prepare_extra_step_kwargs(generator, eta)


        # expand the latents if we are doing classifier free guidance
        latent_model_input = torch.cat([latents] * 2) if do_classifier_free_guidance else latents

        # predict the noise residual
        noise_pred = self.pipeline.unet(
            latent_model_input,
            0,
            encoder_hidden_states=prompt_embeds,
            cross_attention_kwargs=cross_attention_kwargs,
            return_dict=False,
        )[0]

        # perform guidance
        if do_classifier_free_guidance:
            noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
            noise_pred = noise_pred_uncond + guidance_scale * (noise_pred_text - noise_pred_uncond)


        if not output_type == "latent":
            image = self.pipeline.vae.decode(latents / self.pipeline.vae.config.scaling_factor, return_dict=False)[0]
        else:
            image = latents
            has_nsfw_concept = None


class DiffuserModel(AbstractModel):
    def __init__(self, *args, tokenizer=None, **kwargs) -> None:
        self.local_model: Diffuser = None
        self.meta_model: Diffuser = None

        super().__init__(*args, **kwargs)

    def _register_increment_hook(self, hook) -> RemovableHandle:
        return self.local_model.unet.register_forward_hook(hook)

    def _load_meta(self, repoid_or_path, *args, device_map=None, **kwargs) -> torch.nn.Module:
        return Diffuser(
            repoid_or_path, *args, low_cpu_mem_usage=False, device_map=None, **kwargs
        )

    def _load_local(self, repoid_or_path, *args, **kwargs) -> torch.nn.Module:
        return Diffuser(repoid_or_path, *args, **kwargs)

    def _prepare_inputs(
        self,
        inputs,

    ) -> Any:
        return inputs

    def _scan(self, inputs, *args, **kwargs) -> None:
        self.meta_model.scan(inputs, *args, **kwargs)

    def _forward(self, inputs, *args, n_imgs=1, img_size=512, **kwargs) -> None:
        text_tokens, latents = inputs

        text_embeddings = self.meta_model.get_text_embeddings(text_tokens, n_imgs)

        latents = torch.cat([latents] * 2).to("meta")

        return self.meta_model.unet(
            latents,
            torch.zeros((1,), device="meta"),
            encoder_hidden_states=text_embeddings,
        ).sample

    def _generation(
        self,
        inputs,
        *args,
        **kwargs,
    ) -> None:
       
       return self.local_model.pipeline(inputs, *args, **kwargs)

    def _batched_inputs(self, prepared_inputs: BatchEncoding) -> torch.Tensor:
        return prepared_inputs if not isinstance(prepared_inputs, str) else [prepared_inputs]

    def _example_input(self) -> Dict[str, torch.Tensor]:
        return "_"


from __future__ import annotations

from abc import abstractmethod
from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, Any, List

import torch

from ..intervention import InterventionProxy
from ..tracing.Graph import Graph

if TYPE_CHECKING:
    from ..models.AbstractModel import AbstractModel


class Tracer(AbstractContextManager):
    """The Tracer class creates a :class:`nnsight.tracing.Graph.Graph` around the meta_model of a :class:`nnsight.models.AbstractModel.AbstractModel` which tracks and manages the operations performed on the inputs and outputs of said model.

    Attributes:
        model (nnsight.models.AbstractModel.AbstractModel): nnsight Model object that ths context manager traces and executes.
        graph (nnsight.tracing.Graph.Graph): Graph which operations performed on the input and output of Modules are added and later executed.
        args (List[Any]): Positional arguments to be passed to function that executes the model.
        kwargs (Dict[str,Any]): Keyword arguments to be passed to function that executes the model.
        batch_size (int): Batch size of the most recent input. Used by Module to create input/output proxies.
        generation_idx (int): Current generation idx for multi-iteration generation. Used by Module to create input/output proxies.
        batched_input (List[Any]): Batched version of all inputs involved in this Tracer.
        output (Any): Output of execution after __exit__
    """

    def __init__(
        self,
        model: "AbstractModel",
        *args,
        validate:bool = True,
        **kwargs,
    ) -> None:
        self.model = model

        self.args = args
        self.kwargs = kwargs

        self.graph = Graph(self.model.meta_model, proxy_class=InterventionProxy, validate=validate)

        self.batch_size: int = 0
        self.generation_idx: int = 0
        self.batched_input: List[Any] = []

        self.output = None

        # Modules need to know about the current Tracer to create the correct proxies.
        for name, module in self.model.meta_model.named_modules():
            if not isinstance(module, torch.nn.ModuleList):
                module.tracer = self

    @abstractmethod
    def __enter__(self) -> Tracer:
        raise NotImplementedError()

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        raise NotImplementedError()



from nnsight import LanguageModel
from nnsight.tracing.Proxy import Proxy
from nnsight import util

model = LanguageModel('gpt2', device_map='cuda:0')

gpt2_text = "Natural language processing tasks, such as question answering, machine translation, reading comprehension, and summarization, are typically approached with supervised learning on taskspecific datasets."

with model.generate(max_new_tokens=1, output_attentions=True) as generator:

    with generator.invoke(gpt2_text, output_attentions=True) as invoker:

        tokens = invoker.tokens

        attn_hidden_states = [model.transformer.h[layer_idx].attn.output[2][0].save() for layer_idx in range(len(model.transformer.h))]

attn_hidden_states = util.apply(attn_hidden_states, lambda x : x.value, Proxy)

import circuitsvis as cv

cv.attention.attention_patterns(tokens=tokens, attention=attn_hidden_states[0])



# Configuration file for the Sphinx documentation builder.

# Project Information
project = 'nnsight'
copyright = '2023, NDIF'
author = 'Jaden Fiotto-Kaufman'
release = '0.0.7'



# General Configuration
extensions = [
    'sphinx.ext.autodoc', # Auto documentation from docstrings
    'sphinx.ext.napoleon', # Support for NumPy and Google style docstrings
    'sphinx_copybutton', # Copy button for code blocks
    'sphinx_design', # Boostrap design components
    'nbsphinx', # Jupyter notebook support
]

templates_path = ['_templates']
exclude_patterns = []
fixed_sidebar = True



# HTML Output Options

# See https://sphinx-themes.org/ for more
html_theme = "pydata_sphinx_theme"

html_static_path = ['_static']

html_favicon = '_static/images/icon.ico'

html_theme_options = {
  "logo": {"text":"nnsight"},
  "show_nav_level": 2,
  "navbar_end": ["navbar-icon-links", "ndif_status"],
  "navbar_align": "left",
  "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/JadenFiotto-Kaufman/nnsight",
            "icon": "fa-brands fa-github",
        },
  ]
}

html_context = {
   "default_mode": "light",
   "ndif_url" : "https://ndif.baulab.us/ping"
}

html_css_files = [
    'css/custom.css',
]




"""The Module object acts as the entrypoint into the interwoven intervention graph. Through it, one can manipulate the inputs and outputs of Modules as a model is running. Specifically, the ``.output`` and ``.input`` attributes are the root nodes of the intervention graph, and all operations performed on these Proxy objects are the downstream nodes that populate the graph. Requires a ``Tracer`` object to add the correct proxies to the intervention graph. 

    Examples:

        The example below accesses the inputs and outputs of a gpt2 module, saves them during execution, and prints the resulting values:
        
        .. code-block:: python

            with model.generate() as generator:
                with generator.invoke('The Eiffel Tower is in the city of') as invoker:
                    hidden_states = model.lm_head.input.save()
                    logits = model.lm_head.output.save()

            print(hidden_states.value)
            print(logits.value)

        This example shows how to set the output of a gpt2 module to zero:

        .. code-block:: python

            with model.generate() as generator:
                with generator.invoke('The Eiffel Tower is in the city of') as invoker:
                    model.transformer.h[0].output[0] = 0
"""

from __future__ import annotations

from typing import Any, Dict, List, Union

import torch

from . import util
from .contexts.Tracer import Tracer
from .intervention import InterventionProxy
from .tracing.Graph import Graph
from .tracing.Proxy import Proxy


class Module(torch.nn.Module):
    """Class that wraps existing torch modules within a model's module tree in order to add nnsight functionality.
    Proxies of it's output and input are accessed by `.output` and `.input` respectively.

    Attributes:
        module_path (str): String representing the attribute path of this module relative the the root model. Separated by '.' e.x ('transformer.h.0.mlp'). Set by AbstractModel on initialization of meta model.
        output (nnsight.intervention.InterventionProxy): Proxy object representing the output of this module. Reset on pass through.
        output_shape (torch.Size): Shape of the tensor outputs to this module. Populated by most recent pass through. Can also be a nested list of torch.Size.
        output_type (torch.dtype): Dtype of the tensor outputs to this module. Populated by most recent pass through. Can also be a nested list of torch.dtype.
        output (nnsight.intervention.InterventionProxy): Proxy object representing the output of this module. Reset on pass through.
        input_shape (torch.Size): Shape of the tensor inputs to this module. Populated by most recent pass through. Can also be a nested list of torch.Size.
        input_type (torch.dtype): Dtype of the tensor inputs to this module. Populated by most recent pass through. Can also be a nested list of torch.dtype.
        tracer (nnsight.context.Tracer.Tracer): Object which adds this module's output and input proxies to an intervention graph. Must be set on Module objects manually.
    """

    def __init__(self) -> None:
        self.module_path: str = None
        self.input_shape: torch.Size = None
        self.input_type: torch.dtype = None
        self.output_shape: torch.Size = None
        self.output_type: torch.dtype = None

        self._output: InterventionProxy = None
        self._input: InterventionProxy = None
        self._backward_output: InterventionProxy = None
        self._backward_input: InterventionProxy = None
        self._graph: Graph = None

        self.tracer: Tracer = None

    def __call__(
        self, *args: List[Any], **kwds: Dict[str, Any]
    ) -> Union[Any, InterventionProxy]:
        """Override __call__ to check for InterventionProxy arguments.
        If there are any, we should return an InterventionProxy denoting we want to call the given module with arguments.

        Args:
            args (List[Any]): Positional arguments.
            kwargs (Dict[str,Any]): Keyword arguments

        Returns:
            Union[Any,InterventionProxy]: Either the output of the module if not tracing or a Proxy if tracing.
        """
        proxy = any(
            isinstance(x, InterventionProxy) for x in list(args) + list(kwds.values())
        )

        if proxy:
            module_proxy = getattr(self.tracer.graph.module_proxy, self.module_path)

            return module_proxy.forward(*args, **kwds)

        return super().__call__(*args, **kwds)

    @property
    def output(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the output of this module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            Proxy: Output proxy.
        """
        if self._output is None:
            self._output = self.tracer.graph.add(
                value=util.apply(
                    self.output_shape,
                    lambda x: torch.empty(x, device="meta", requires_grad=True).clone(),
                    torch.Size,
                ),
                target="argument",
                args=[
                    f"{self.module_path}.output.{self.tracer.generation_idx}",
                    self.tracer.batch_size,
                    len(self.tracer.batched_input) - self.tracer.batch_size,
                ],
            )

        return self._output

    @output.setter
    def output(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the output of this module and therefore we create a Proxy of that request.

        Args:
            value (Union[Proxy, Any]): Value to set output to.
        """

        self.output.node.graph.add(
            target=Proxy.proxy_update,
            args=[self.output.node, value],
        )

    @property
    def input(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the input of this module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            Proxy: Input proxy.
        """
        if self._input is None:
            self._input = self.tracer.graph.add(
                value=util.apply(
                    self.input_shape,
                    lambda x: torch.empty(x, device="meta", requires_grad=True).clone(),
                    torch.Size,
                ),
                target="argument",
                args=[
                    f"{self.module_path}.input.{self.tracer.generation_idx}",
                    self.tracer.batch_size,
                    len(self.tracer.batched_input) - self.tracer.batch_size,
                ],
            )

        return self._input

    @input.setter
    def input(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the input of this module and therefore we create a Proxy of that request.

        Args:
            value (Union[Proxy, Any]): Value to set input to.
        """

        self.input.node.graph.add(
            target=Proxy.proxy_update,
            args=[self.input.node, value],
        )

    @property
    def backward_output(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the backward_output of this module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            Proxy: backward_output proxy.
        """
        if self._backward_output is None:
            self._backward_output = self.tracer.graph.add(
                value=util.apply(
                    self.output_shape,
                    lambda x: torch.empty(x, device="meta"),
                    torch.Size,
                ),
                target="argument",
                args=[
                    f"{self.module_path}.backward_output.{self.tracer.generation_idx}",
                    self.tracer.batch_size,
                    len(self.tracer.batched_input) - self.tracer.batch_size,
                ],
            )

        return self._backward_output

    @backward_output.setter
    def backward_output(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the backward_output of this module and therefore we create a Proxy of that request.

        Args:
            value (Union[Proxy, Any]): Value to set backward_output to.
        """

        self.backward_output.node.graph.add(
            target=Proxy.proxy_update,
            args=[self.backward_output.node, value],
        )

    @property
    def backward_input(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the backward_input of this module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            Proxy: backward_input proxy.
        """
        if self._backward_input is None:
            self._backward_input = self.tracer.graph.add(
                value=util.apply(
                    self.input_shape,
                    lambda x: torch.empty(x, device="meta"),
                    torch.Size,
                ),
                target="argument",
                args=[
                    f"{self.module_path}.backward_input.{self.tracer.generation_idx}",
                    self.tracer.batch_size,
                    len(self.tracer.batched_input) - self.tracer.batch_size,
                ],
            )

        return self._backward_input

    @backward_input.setter
    def backward_input(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the backward_input of this module and therefore we create a Proxy of that request.

        Args:
            value (Union[Proxy, Any]): Value to set backward_input to.
        """

        self.backward_input.node.graph.add(
            target=Proxy.proxy_update,
            args=[self.backward_input.node, value],
        )

    @property
    def graph(self) -> Graph:
        if self._graph is None:
            self._graph = Graph.trace(
                self,
                *util.apply(
                    self.input_shape,
                    lambda x: torch.empty(x, device="meta"),
                    torch.Size,
                ),
            )

        return self._graph

    @staticmethod
    def wrap(module: torch.nn.Module) -> Module:
        """Wraps the torch Module with our Module class. Adds a forward hook to the module to populate output and input data. Does not wrap torch.nn.ModuleList.

        Args:
            module (torch.nn.Module): The torch Module to wrap.

        Returns:
            Module: The wrapped Module.
        """

        def hook(module: Module, input: Any, output: Any):
            module._output = None
            module._input = None
            module._backward_output = None
            module._backward_input = None
            module.output_shape = util.apply(output, lambda x: x.shape, torch.Tensor)
            module.input_shape = util.apply(input, lambda x: x.shape, torch.Tensor)
            module.output_type = util.apply(output, lambda x: x.dtype, torch.Tensor)
            module.input_type = util.apply(input, lambda x: x.dtype, torch.Tensor)

        for name, _module in module.named_children():
            setattr(module, name, Module.wrap(_module))

        if isinstance(module, (Module, torch.nn.ModuleList)):
            return module

        util.wrap(module, Module)

        module.register_forward_hook(hook)

        return module



from __future__ import annotations

import operator
from typing import TYPE_CHECKING, Any, Callable, Union

import torch

from .. import util

if TYPE_CHECKING:
    from .Node import Node


class Proxy:
    """Proxy objects are the actual objects that interact with operations in order to update the graph to create new nodes.

    The operations that are traceable on base Proxy objects are many python built-in and magic methods, as well as implementing __torch_function__ to trace torch operations.

    Attributes:
        node (Node): This proxy's node.
    """

    @staticmethod
    def proxy_update(value1: Any, value2: Any) -> None:
        """Updates Tensor values with other Tensor values.

        Args:
            value1 (Any): Collection with Tensors to update.
            value2 (Any): Collection with Tensors to pull values from.
        """
        if isinstance(value1, torch.Tensor):
            value1[:] = value2
        elif isinstance(value1, list) or isinstance(value1, tuple):
            for value_idx in range(len(value1)):
                Proxy.proxy_update(value1[value_idx], value2[value_idx])
        elif isinstance(value1, dict):
            for key in value1:
                Proxy.proxy_update(value1[key], value2[key])

    @staticmethod
    def proxy_call(callable: Callable, *args, **kwargs) -> None:
        return callable(*args, **kwargs)

    def __init__(self, node: "Node") -> None:
        self.node = node

    def __call__(self, *args, **kwargs) -> Proxy:
        """
        Calling a Proxy object normally just creates a Proxy.proxy_call operation. However if this call is a method on the root module proxy, it's assumed that one wishes to trace into the method and therefore trace all operations inside it.

        Returns:
            Proxy: New call proxy.
        """
        # If calling a method (not a sub-module) on the main module of this graph,
        # we want to trace into that method.
        if self.node.args[0] is self.node.graph.module_proxy.node and not isinstance(
            self.node.proxy_value, torch.nn.Module
        ):
            value = self.node.proxy_value.__func__(
                self.node.graph.module_proxy, *args, **kwargs
            )

            return value
        # Otherwise we just want to add a node saying we wish to call this module.
        else:
            return self.node.graph.add(
                target=Proxy.proxy_call,
                args=[self.node] + list(args),
                kwargs=kwargs,
            )

    def __getitem__(self, key: Union[Proxy, Any]) -> Proxy:
        return self.node.graph.add(
            target=operator.getitem,
            args=[self.node, key],
        )

    def __setitem__(self, key: Union[Proxy, Any], value: Union[Proxy, Any]) -> None:
        item_proxy = self[key]

        item_proxy.node.graph.add(
            target=Proxy.proxy_update,
            args=[item_proxy.node, value],
        )

    def __getattr__(self, key: Union[Proxy, Any]) -> Proxy:
        return self.node.graph.add(
            target=util.fetch_attr,
            args=[self.node, key],
        )

    def __len__(self) -> Proxy:
        return self.node.graph.add(
            target=len,
            args=[self.node],
        )

    def __add__(self, other: Union[Proxy, Any]) -> Proxy:
        return self.node.graph.add(
            target=operator.add,
            args=[self.node, other],
        )

    def __sub__(self, other: Union[Proxy, Any]) -> Proxy:
        return self.node.graph.add(
            target=operator.sub,
            args=[self.node, other],
        )

    def __pow__(self, other: Union[Proxy, Any]) -> Proxy:
        return self.node.graph.add(
            target=pow,
            args=[self.node, other],
        )

    def __mul__(self, other: Union[Proxy, Any]) -> Proxy:
        return self.node.graph.add(
            target=operator.mul,
            args=[self.node, other],
        )

    def __matmul__(self, other: Union[Proxy, Any]) -> Proxy:
        return self.node.graph.add(
            target=operator.matmul,
            args=[self.node, other],
        )

    def __truediv__(self, other: Union[Proxy, Any]) -> Proxy:
        return self.node.graph.add(
            target=operator.truediv,
            args=[self.node, other],
        )

    def __bool__(self) -> bool:
        return self.node.proxy_value.__bool__()

    def __index__(self) -> int:
        return self.node.proxy_value.__index__()

    def __instancecheck__(self, __instance: Any) -> bool:
        return self.node.proxy_value.__instancecheck__(__instance)

    @classmethod
    def __torch_function__(cls, orig_method, types, args=None, kwargs=None) -> Proxy:
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()

        proxy: Proxy = None

        for arg in args:
            if isinstance(arg, Proxy):
                proxy = arg
                break

        return proxy.node.graph.add(
            target=orig_method,
            args=args,
            kwargs=kwargs,
        )


from functools import wraps


def proxy_wrapper(fn) -> None:
    """Wraps problematic functions (torch functions sometimes).
    Checks if any of its args are proxies. If so we return a proxy of the function.
    Otherwise just run the function.

    Args:
        fn (function): Function to wrap.

    Returns:
        function: Wrapped function.
    """

    @wraps(fn)
    def patched(*args, **kwargs):
        arguments = list(args) + list(kwargs.values())

        node = None

        for arg in arguments:
            if isinstance(arg, Proxy):
                node = arg.node

                break

        if node is not None:
            value = fn(
                *node.prepare_proxy_values(args),
                **node.prepare_proxy_values(kwargs),
            )

            return node.graph.add(value=value, target=fn, args=args, kwargs=kwargs)

        else:
            return fn(*args, **kwargs)

    return patched



from nnsight import LanguageModel

model = LanguageModel('gpt2', device_map='cuda:0')

with model.generate(max_new_tokens=3) as generator:
    
    with generator.invoke("Madison square garden is located in the city of New") as invoker:

        embeddings = model.transformer.wte.output

print(generator.output)

model = LanguageModel(model.local_model, tokenizer=model.tokenizer)

with model.generate(max_new_tokens=3) as generator:
    
    with generator.invoke("Madison square garden is located in the city of New") as invoker:

        embeddings = model.transformer.wte.output

print(generator.output)



from nnsight import LanguageModel

model = LanguageModel("gpt2", device_map='cuda:0')

with model.forward(inference=False) as runner: # <-- inference mode is turned off
    with runner.invoke("testing") as invoker:
        
        logits = model.lm_head.output

        mlp = model.transformer.h[0].mlp.output.save() # <-- get mlp activations
        mlp.retain_grad() # <-- tell activations to retain it's grad

        loss = logits.sum()
        loss.backward() # <-- do backward

print(mlp.value.grad) # <-- grad is populated

with model.forward(inference=False) as runner: # <-- inference mode is turned off
    with runner.invoke("testing") as invoker:
        
        logits = model.lm_head.output

        mlp_grad = model.transformer.h[0].mlp.backward_output.save() # <-- Directly get grad

        loss = logits.sum()
        loss.backward() # <-- do backward

print(mlp_grad.value) # <-- value is populated



        


from nnsight import LanguageModel

model = LanguageModel('gpt2', device_map='cuda:0')

print(model.transformer.h[1].attn.graph)

model.modulize(model.transformer.h[1].attn, 'softmax_0', 'attention_probs')

with model.generate(max_new_tokens=3) as generator:
    
    with generator.invoke('Hello world') as invoker:

        attention_probs = model.transformer.h[1].attn.attention_probs.output.save()
    
print(attention_probs.value)




from typing import Any

import torch
from nnsight import AbstractModel, LanguageModel, util
from nnsight.module import Module
from torch.utils.data import DataLoader, Dataset

model = LanguageModel("gpt2", device_map="cuda:0")

n_tokens = 10
epochs = 1
answer = "Paris"
answer_tokens = model.tokenizer(answer)
answer_token = answer_tokens["input_ids"][0]


class SoftPrompt:
    def __init__(self, module: Module, n: int) -> None:
        self.module = module
        self.n = n

        self.embedding = torch.nn.Parameter(
            torch.zeros((self.n, self.module.embedding_dim)), requires_grad=True
        )

    def __call__(self) -> Any:
        self.module.output = self.embedding[:]

    def parameters(self):
        return [self.embedding]


sp = SoftPrompt(model.transformer.wte, n_tokens)

optimizer = torch.optim.AdamW(sp.parameters())
dataset = [[" ".join(["_"] * n_tokens), answer_token]] * 100
dataloader = DataLoader(dataset, batch_size=10)


def decode(hs):
    return torch.log_softmax(model.lm_head(model.transformer.ln_f(hs)), dim=-1)

with model.generate(max_new_tokens=2) as generator:

    with generator.invoke("Madison Square Garden is located in New") as invoker:

        hs = model.transformer.h[1].output[0].t[-1].save()

        invoker.next()

        target = decode(model.transformer.h[-1].output[0]).save()

target = target.value
hs = hs.value

lossfn = torch.nn.KLDivLoss(reduction="batchmean", log_target=True)

for epoch in range(epochs):
    print(epoch)

    for i, (inputs, targets) in enumerate(dataloader):
        print(f"  {i}")

        optimizer.zero_grad()

        with model.forward(inputs, inference=False) as runner:

            sp()

            model.transformer.h[1].output[0].t[-1] = hs

            pred = decode(model.transformer.h[-1].output[0]).save()



        loss = lossfn(pred.value, target)
        loss.backward()

        optimizer.step()


with model.generate() as generator:
    with generator.invoke(dataset[0][0]) as invoker:
        pass

print(model.tokenizer.decode(generator.output[0]))


with model.generate() as generator:
    with generator.invoke(dataset[0][0]) as invoker:
        sp()

print(model.tokenizer.decode(generator.output[0]))



from __future__ import annotations

import copy
import gc
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Union

import accelerate
import torch
from torch.utils.hooks import RemovableHandle

from ..alteration import REPOID_TO_ALTERATION
from ..contexts.DirectInvoker import DirectInvoker
from ..contexts.Runner import Runner
from ..editing.Editor import Edit, Editor
from ..editing.GraphEdit import GraphEdit
from ..editing.WrapperModuleEdit import WrapperModuleEdit
from ..intervention import HookModel, intervene
from ..logger import logger
from ..module import Module
from ..patching import Patch, Patcher
from ..tracing.Graph import Graph


class AbstractModel(ABC):
    """Abstract class to be implemented for PyTorch models wishing to gain this package's functionality.

    Attributes:
        repoid_path_clsname (str): Hugging face repo id of model to load, path to checkpoint, or class name of custom model.
        args (List[Any]): Positional arguments used to initialize model.
        kwargs (Dict[str,Any]): Keyword arguments used to initialize model.
        alter (bool): If to check for alterations and apply them to the model. Defaults to True.
        dispatched (bool): If the local_model has bee loaded yet.
        dispatch (bool): If to load and dispatch model on init. Defaults to False.
        custom_model (bool): If the value passed to repoid_path_model was a custom model.
        meta_model (nnsight.Module): Version of the root model where all parameters and tensors are on the 'meta'
            device. All modules are wrapped in nnsight.Module adding interleaving operation functionality.
        local_model (torch.nn.Module): Locally loaded and dispatched model. Only loaded and dispatched on first use.
            This is the actual model that is ran with hooks added to it to enter the intervention graph.
    """

    def __init__(
        self,
        repoid_path_model: Union[str, torch.nn.Module],
        *args,
        dispatch: bool = False,
        alter: bool = True,
        **kwargs,
    ) -> None:
        super().__init__()

        self.repoid_path_clsname = repoid_path_model
        self.args = args
        self.kwargs = kwargs
        self.alter = alter
        self.dispatch = dispatch
        self.dispatched = False
        self.custom_model = False
        self.meta_model: Module = None
        self.local_model: torch.nn.Module = None
        self.edits: List[Edit] = list()

        # If using a custom passed in model.
        if isinstance(repoid_path_model, torch.nn.Module):
            self.repoid_path_clsname = repoid_path_model.__class__.__name__
            self.custom_model = True
            self.dispatched = True
            self.local_model = repoid_path_model

        logger.info(f"Initializing `{self.repoid_path_clsname}`...")

        # If alter and alteration exist, use alteration patcher context while loading module.
        with self.alteration() if self.alter else Patcher():
            # Use accelerate and .to('meta') to assure tensors are loaded to 'meta' device
            with accelerate.init_empty_weights(include_buffers=True):
                if self.custom_model:
                    # Need to force parameters when deepcopied, to instead create a meta tensor of the same size/dtype
                    def meta_deepcopy(self, memo):
                        if id(self) in memo:
                            return memo[id(self)]
                        else:
                            result = type(self)(
                                torch.empty_like(
                                    self.data, dtype=self.data.dtype, device="meta"
                                ),
                                self.requires_grad,
                            )
                            memo[id(self)] = result
                            return result

                    # Patching Parameter __deepcopy__
                    with Patcher() as patcher:
                        patcher.add(
                            Patch(
                                torch.nn.parameter.Parameter.__deepcopy__, meta_deepcopy
                            )
                        )

                        self.meta_model: Module = Module.wrap(
                            copy.deepcopy(self.local_model).to("meta")
                        )
                else:
                    self.meta_model: Module = Module.wrap(
                        self._load_meta(self.repoid_path_clsname, *args, **kwargs).to(
                            "meta"
                        )
                    )

        # Wrap all modules in our Module class.
        for name, module in self.meta_model.named_children():
            module = Module.wrap(module)

            setattr(self.meta_model, name, module)

        # Set module_path attribute so Modules know their place.
        for name, module in self.meta_model.named_modules():
            module.module_path = name

        # Run initial dummy string to populate Module shapes, dtypes etc
        self._scan(self._prepare_inputs(self._example_input()))

        if self.dispatch:
            self.dispatch_local_model()

        logger.info(f"Initialized `{self.repoid_path_clsname}`")

    def __repr__(self) -> str:
        return repr(self.meta_model)

    def __getattr__(self, key: Any) -> Any:
        """Allows access of sub-modules on meta_model directly from AbstractModel object

        Args:
            key (Any): Key.

        Returns:
            Any: Attribute.
        """
        return getattr(self.meta_model, key)

    def __call__(
        self,
        fn: Callable,
        inputs: Any,
        graph: Graph,
        *args,
        edits: List[Edit] = None,
        inference: bool = True,
        **kwargs,
    ) -> Any:
        """Runs some function with some inputs and some graph with the appropriate contexts for this model.

        Loads and dispatched local_model if not already done so.

        Args:
            fn (Callable): Function to run.
            inputs (Any): Inputs to give to function.
            graph (Graph): Intervention graph to interleave with model's computation graph.
            inference (bool): If running in inference mode. Defaults to True.

        Returns:
            Any: Output of model.
        """
        if edits is None:
            edits = self.edits

        # If local_model not yet loaded, do so.
        if not self.dispatched:
            self.dispatch_local_model()

        with Editor(self, edits):
            increment_hook = self._register_increment_hook(
                lambda module, input, output: graph.increment()
            )

            logger.info(f"Running `{self.repoid_path_clsname}`...")

            # Send local_model to graph to re-compile
            graph.compile(self.local_model)

            inputs = self._prepare_inputs(inputs)

            with torch.inference_mode(mode=inference):
                with HookModel(
                    self.local_model,
                    list(graph.argument_node_names.keys()),
                    input_hook=lambda activations, module_path: intervene(
                        activations, module_path, graph, "input"
                    ),
                    output_hook=lambda activations, module_path: intervene(
                        activations, module_path, graph, "output"
                    ),
                    backward_input_hook=lambda activations, module_path: intervene(
                        activations, module_path, graph, "backward_input"
                    ),
                    backward_output_hook=lambda activations, module_path: intervene(
                        activations, module_path, graph, "backward_output"
                    ),
                ):
                    output = fn(inputs, *args, **kwargs)

            increment_hook.remove()

            logger.info(f"Completed `{self.repoid_path_clsname}`")

            gc.collect()
            torch.cuda.empty_cache()

        return output

    def dispatch_local_model(self, *args, **kwargs) -> None:
        with self.alteration() if self.alter else Patcher():
            self.local_model = self._load_local(
                self.repoid_path_clsname, *self.args, *args, **kwargs, **self.kwargs
            )

        self.dispatched = True

    def generate(self, *args, **kwargs) -> Runner:
        """Returns a Runner context for this model's _generation method.

        Runner contexts are used to trace and interleave operations on the model's computation graph.

        Arguments passed to ``.generate()`` are passed downstream to the model specific _generation method.

        Runners are used in tandem with their Invoker contexts to enter inputs for operation tracing and execution.

        The output of the run is ultimately saved in the generator's ``.output`` attribute.

        Returns:
            Runner: Runner.

        Examples:

            A simple entering of a generation context on a language model, and running a prompt with no interventions:

            .. code-block:: python

                with model.generate(max_new_tokens=1) as generator:
                    with generator.invoke('The Eiffel Tower is in the city of') as invoker:
                        pass

                print(generator.output)

            Keyword arguments like 'max_new_tokens' are model specific and in this case limits the amount of tokens to predict to one.

            See the Runner docs for more information.

        """
        return Runner(self, *args, **kwargs, generation=True)

    def forward(self, *args, **kwargs) -> Runner:
        """Returns a Runner context for this model's _forward method.

        Runner contexts are used to trace and interleave operations on the model's computation graph.

        Arguments passed to ``.forward()`` are passed downstream to the model specific _forward method.

        Runners are used in tandem with their Invoker contexts to enter inputs for operation tracing and execution.

        The output of the run is ultimately saved in the runner's ``.output`` attribute.

        Returns:
            Runner: Runner.

        Examples:

            A simple entering of a forward context on a language model, and running a prompt with no interventions:

            .. code-block:: python

                with model.forward() as runner:
                    with runner.invoke('The Eiffel Tower is in the city of') as invoker:
                        pass

                print(runner.output)

            See the Runner docs for more information.

        """
        return Runner(self, *args, **kwargs, generation=False)

    def invoke(self, *args, **kwargs):
        """Returns a Runner context for this model's _forward method. Also acts as an invocation for the runner allowing you to create a Runner anf Invocation context in one go.
        Gives option to create Runner context and call an invoke on it in one call. Adds an extra 'fwd_args' kwarg to allow the args normally passed to ``.forward()``
        Returns:
            Runner: Runner.

        Examples:

            A simple entering of a forward context on a language model, and running a prompt with no interventions:

            .. code-block:: python

                with model.invoke('The Eiffel Tower is in the city of', fwd_args={}) as invoker:
                    pass

                print(runner.output)

            See the Runner docs for more information.

        """
        return DirectInvoker(self, *args, **kwargs)

    def alteration(self) -> Patcher:
        return REPOID_TO_ALTERATION.get(self.repoid_path_clsname, Patcher())

    def modulize(self, module: Module, node_name: str, module_name: str) -> None:
        """_summary_

        Args:
            module (Module): _description_
            node_name (str): _description_
            module_name (str): _description_
        """

        # Create a WrapperModuleEdit which just adds a WrapperModule to an existing module at the given module_name.
        wme = WrapperModuleEdit(module.module_path, module_name)
        # Wrap with our Module and update new attributes.
        wme.wrapper: Module = Module.wrap(wme.wrapper)
        wme.wrapper.module_path = f"{module.module_path}.{module_name}"
        wme.wrapper.tracer = module.tracer
        wme.wrapper.output_shape = module.output_shape
        # Carry out the edit on the meta_model.
        wme.edit(self.meta_model)

        # Get/create the execution graph for the module's forward method.
        graph = module.graph

        # Add two proxies/nodes, one to get the new WrapperModule we added and another to call it with the data from the original module.
        # Passing the data through the wrapper module allows hooking of the module's output like usual.
        module_proxy = getattr(graph.module_proxy, module_name)
        module_proxy(graph.nodes[node_name])

        # Create and carry out the edit on the meta_model.
        ge = GraphEdit(module.module_path, module.graph)
        ge.edit(self.meta_model)

        # Append to self.edits so when we call the local model, we temporarily edit the module in the same way as the meta model.
        self.edits.append(wme)
        self.edits.append(ge)

    @abstractmethod
    def _prepare_inputs(self, inputs: Any, **kwargs) -> Any:
        """Abstract method which prepares inputs. To be implemented by inheritors.

        Args:
            inputs (Any): Inputs.

        Returns:
            Any: Prepared inputs.
        """
        raise NotImplementedError()

    @abstractmethod
    def _load_meta(self, repoid_or_path: str, *args, **kwargs) -> torch.nn.Module:
        """
        Abstract method to initialize meta_model. To be implemented by inheritors.

        Args:
            repoid_or_path (str): Huggingface repo id or path to checkpoint.

        Returns:
            torch.nn.Module: Meta version of model
        """
        raise NotImplementedError()

    @abstractmethod
    def _load_local(self, repoid_or_path, *args, **kwargs) -> torch.nn.Module:
        """
        Abstract method to initialize and dispatch the local_model. To be implemented by inheritors.

        Args:
            repoid_or_path (str): Huggingface repo id or path to checkpoint.

        Returns:
            torch.nn.Module: Local version of model
        """
        raise NotImplementedError()

    @abstractmethod
    def _scan(self, inputs, *args, **kwargs) -> None:
        """
        Abstract method to directly call the meta_model and therefore populate the input/output shapes etc. To be implemented by inheritors.

        Used for tracing operations and their input/output shapes/dtypes.

        Args:
            inputs (Any): Inputs.
        """
        raise NotImplementedError()

    @abstractmethod
    def _forward(self, inputs, *args, **kwargs) -> Any:
        """
        Abstract method to directly call the local_model. To be implemented by inheritors.

        Args:
            inputs (Any): Inputs.

        Returns:
            Any: Output.
        """
        raise NotImplementedError()

    @abstractmethod
    def _generation(self, inputs, *args, **kwargs) -> Any:
        """
        Abstract method to do iterative generation on the local_model. To be implemented by inheritors.

        Args:
            inputs (Any): Inputs.

        Returns:
            Any: Output.
        """
        raise NotImplementedError()

    @abstractmethod
    def _register_increment_hook(self, hook: Callable) -> RemovableHandle:
        """Abstract method to hook a function on the local_model on the main module that is incremented during generation.

        Args:
            hook (Callable): Increment hook. Probably from the Generator context.

        Returns:
            RemovableHandle: Handle to remove the applied hook after generation is done.
        """
        raise NotImplementedError()

    @abstractmethod
    def _example_input(self) -> Any:
        """Abstract to provide an example input to be used with ``._scan(...)``

        Returns:
            Any: Example input.
        """
        raise NotImplementedError()

    @abstractmethod
    def _batched_inputs(self, prepared_inputs: Any) -> List[Any]:
        """Abstract to return a version of the prepared inputs from ``._prepare_inputs(...)`` that can be batched with others.
        For example with a LanguageModel, prepare_inputs returns a dictionary. The implementation for this method just returns the 'input_ids'.

        Args:
            prepared_inputs (Any): Inputs from ``._prepare_inputs(...)``

        Returns:
            List[Any]: Batched version of prepared_inputs.
        """
        raise NotImplementedError()



from nnsight import LanguageModel
import torch

model = LanguageModel("gpt2", device_map='cuda:0')

with model.generate() as generator:
    with generator.invoke('The Eiffel Tower is in the city of') as invoker:
        
        hidden_states = model.transformer.h[-1].output[0]
        hidden_states = model.lm_head(model.transformer.ln_f(hidden_states)).save()
        tokens = torch.softmax(hidden_states, dim=2).argmax(dim=2).save()
        
print(hidden_states.value)
print(tokens.value)
print(model.tokenizer.decode(tokens.value[0]))



from nnsight import LanguageModel

model = LanguageModel("gpt2", device_map="cuda:0")


with model.invoke("hello world") as invoker:
    hidden_states = model.transformer.h[-1].output[0].save()


print(invoker.output)
print(hidden_states.value)



import os

import yaml

from .patching import *
from .pydantics.Config import ConfigModel

PATH = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(PATH, "config.yaml"), "r") as file:
    CONFIG = ConfigModel(**yaml.safe_load(file))

from .models.AbstractModel import AbstractModel
from .models.DiffuserModel import DiffuserModel
from .models.LanguageModel import LanguageModel
from .module import Module
from .patching import Patch, Patcher

# Below do default patching:
DEFAULT_PATCHER = Patcher()

from functools import wraps

import torch


# Need to patch repeat_interleave to work with meta tensors
# Computes appropriate shape if meta. Otherwise just call repeat_interleave
def repeat_interleave_wrapper(fn):
    @wraps(fn)
    def repeat_interleave(
        input: torch.Tensor, repeats: torch.LongTensor, dim=None, output_size=None
    ):
        if input.device.type == "meta":
            if not isinstance(repeats, torch.Tensor):
                repeats = torch.LongTensor([repeats])

            if dim is None:
                input = input.flatten()
                dim = 0

            if repeats.dim() == 0 or (repeats.dim() == 1 and repeats.size(0) == 1):
                repeats = repeats.reshape([1]).expand([input.size(dim)])

            new_dim_size = repeats.cumsum(0)[-1].item()
            new_output_shape = list(input.shape)
            new_output_shape[dim] = new_dim_size

            return torch.empty(new_output_shape, device="meta")

        else:
            return fn(input, repeats, dim=dim, output_size=output_size)

    return repeat_interleave


DEFAULT_PATCHER.add(
    Patch(torch.repeat_interleave, repeat_interleave_wrapper(torch.repeat_interleave))
)


DEFAULT_PATCHER.__enter__()

from torch._meta_registrations import (_meta_lib_dont_use_me_use_register_meta,
                                       aten, global_decomposition_table,
                                       register_meta)


# Function which "activates" the most recent meta registered function.
def activate_recent_meta():
    op_overload, fn = list(global_decomposition_table["meta"].items())[-1]
    op_overload.py_impl(torch._C.DispatchKey.Meta)(fn)
    _meta_lib_dont_use_me_use_register_meta.impl(op_overload, fn)


# Need to patch local_scalar_dense_meta for meta.
# In non-meta tensors this returns the singular value in tensors with one value
# In meta, lets just return 0
@register_meta(aten._local_scalar_dense)
def local_scalar_dense_meta(A):
    return 0


activate_recent_meta()



from nnsight import LanguageModel

model = LanguageModel('gpt2', device_map='cuda:0')

print(model)

with model.generate(max_new_tokens=3) as generator:
    
    with generator.invoke("Madison square garden is located in the city of New") as invoker:

        embeddings = model.transformer.wte.output

    with generator.invoke("_ _ _ _ _ _ _ _ _ _") as invoker:

        model.transformer.wte.output = embeddings

print(model.tokenizer.decode(generator.output[0]))
print(model.tokenizer.decode(generator.output[1]))


from __future__ import annotations

import pickle

import socketio

from .. import CONFIG, pydantics
from .Invoker import Invoker
from .Tracer import Tracer


class Runner(Tracer):
    """The Runner object manages the intervention tracing for a given model's _generation method or _run_local method.

    Examples:

        Below example shows accessing the input and output of a gpt2 module, saving them during execution, and printing the resulting values after execution:

        .. code-block:: python

            with model.generate() as generator:
                with generator.invoke('The Eiffel Tower is in the city of') as invoker:
                    hidden_states = model.lm_head.input.save()
                    logits = model.lm_head.output.save()

            print(hidden_states.value)
            print(logits.value)

        Below example shows accessing the output of a gpt2 module and setting the values to zero:

        .. code-block:: python

            with model.forward() as runner:
                with runner.invoke('The Eiffel Tower is in the city of') as invoker:
                    model.transformer.h[0].output[0] = 0

    Attributes:
        generation (bool): If to use the _generation method of the model. Otherwise the _run_local method
        remote (bool): If to use the remote NDIF server for execution of model and computation graph. (Assuming it's running/working)
        blocking (bool): If when using the server option, to hang until job completion or return information you can use to retrieve the job result.
    """

    def __init__(
        self,
        *args,
        generation: bool = False,
        blocking: bool = True,
        remote: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.generation = generation
        self.remote = remote
        self.blocking = blocking

    def __enter__(self) -> Runner:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """On exit, run and generate using the model whether locally or on the server."""
        if isinstance(exc_val, BaseException):
            raise exc_val
        if self.remote:
            self.run_server()
        else:
            self.run_local()

    def run_local(self):
        """Runs the local_model using it's chosen method."""
        # Run the model and store the output.
        self.output = self.model(
            self.model._generation if self.generation else self.model._forward,
            self.batched_input,
            self.graph,
            *self.args,
            **self.kwargs,
        )

    def run_server(self):
        # Create the pydantic class for the request.
        request = pydantics.RequestModel(
            args=self.args,
            kwargs=self.kwargs,
            model_name=self.model.repoid_path_clsname,
            batched_input=self.batched_input,
            intervention_graph=self.graph,
            generation=self.generation,
        )

        if self.blocking:
            self.blocking_request(request)
        else:
            self.non_blocking_request(request)

    def blocking_request(self, request: pydantics.RequestModel):
        # Create a socketio connection to the server.
        sio = socketio.Client()
        sio.connect(f"ws://{CONFIG.API.HOST}", transports=['websocket'])

        # Called when receiving a response from the server.
        @sio.on("blocking_response")
        def blocking_response(data):
            # Load the data into the ResponseModel pydantic class.
            data: pydantics.ResponseModel = pickle.loads(data)

            # Print response for user ( should be logger.info and have an info handler print to stdout)
            print(str(data))

            # If the status of the response is completed, update the local nodes that the user specified to save.
            # Then disconnect and continue.
            if data.status == pydantics.JobStatus.COMPLETED:
                for name, value in data.saves.items():
                    self.graph.nodes[name].value = value

                self.output = data.output

                sio.disconnect()
            # Or if there was some error.
            elif data.status == pydantics.JobStatus.ERROR:
                sio.disconnect()

        sio.emit(
            "blocking_request",
            request.model_dump(exclude_defaults=True, exclude_none=True),
        )

        sio.wait()

    def non_blocking_request(self, request: pydantics.RequestModel):
        pass

    def invoke(self, input, *args, **kwargs) -> Invoker:
        return Invoker(self, input, *args, **kwargs)
